# Copyright 2022 MOSEC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The Coordinator is used to control the data flow between `Worker` and `Server`."""

import enum
import os
import queue
import signal
import socket
import struct
import threading
import time
import traceback
from contextlib import contextmanager
from multiprocessing.synchronize import Event
from typing import Any, Optional, Sequence, Type

from mosec.errors import MosecError, MosecTimeoutError
from mosec.log import get_internal_logger
from mosec.protocol import HTTPStautsCode, Protocol
from mosec.worker import SSEWorker, Worker

logger = get_internal_logger()


CONN_MAX_RETRY = 10
CONN_CHECK_INTERVAL = 1


class State(enum.IntFlag):
    """Task state."""

    INGRESS = 0b1
    EGRESS = 0b10


PROTOCOL_TIMEOUT = 2.0


@contextmanager
def set_mosec_timeout(duration: int):
    """Context manager to set a timeout for a code block.

    Args:
        duration (float): the duration in seconds before timing out
    """

    def handler(signum, frame):
        raise MosecTimeoutError(
            f"[{signum}]`forward` timeout after {duration}s: {frame}"
        )

    signal.signal(signal.SIGALRM, handler)
    signal.alarm(duration)
    try:
        yield
    finally:
        signal.alarm(0)


class FakeSemaphore:
    """Fake semaphore interface for non-SSE workers."""

    def __init__(self, value: int = 1):
        """Not intend to be used by users."""
        self.value = value

    def acquire(self, timeout: Optional[float] = None):
        """Acquire the semaphore."""

    def release(self):
        """Release the semaphore."""

    __enter__ = acquire

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context."""
        self.release()


class Coordinator:
    """Coordinator controls the data flow.

    This private class defines a set of coordination behaviors
    to receive tasks via `Protocol` and process them via `Worker`.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        worker: Type[Worker],
        max_batch_size: int,
        shutdown: Event,
        shutdown_notify: Event,
        socket_prefix: str,
        stage_name: str,
        worker_id: int,
        timeout: int,
    ):
        """Initialize the mosec coordinator.

        Args:
            worker: subclass of `mosec.Worker` implemented by users.
            max_batch_size: maximum batch size for this worker.
            shutdown: `multiprocessing.synchronize.Event` object for shutdown
                IPC.
            socket_prefix: prefix for the socket addresses.
            stage_name: identification name for this worker stage.
            worker_id: identification number for worker processes at the same
                stage.
            timeout: timeout for the `forward` function.
        """
        self.name = f"<{stage_name}|{worker_id}>"
        self.timeout = timeout
        self.current_ids: Sequence[bytes] = []
        self.semaphore = FakeSemaphore()  # type: ignore

        self.worker = worker()
        self.worker.worker_id = worker_id
        self.worker.max_batch_size = max_batch_size
        self.worker.stage = stage_name

        self.protocol = Protocol(
            name=self.name,
            addr=os.path.join(socket_prefix, f"ipc_{stage_name}.socket"),
            timeout=PROTOCOL_TIMEOUT,
        )

        # ignore termination & interruption signal
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        self.shutdown = shutdown
        self.shutdown_notify = shutdown_notify

        # SSE support
        if issubclass(worker, SSEWorker):
            self.stream: queue.SimpleQueue = queue.SimpleQueue()
            self.semaphore: threading.Semaphore = threading.Semaphore()  # type: ignore
            self.worker._stream_queue = self.stream  # type: ignore
            self.worker._stream_semaphore = self.semaphore  # type: ignore
            threading.Thread(target=self.streaming, daemon=True).start()

        self.warmup()
        self.init_protocol()
        self.run()

    def init_protocol(self):
        """Check socket readiness."""
        retry_count = 0
        while retry_count < CONN_MAX_RETRY and not self.shutdown.is_set():
            if os.path.exists(self.protocol.addr):
                return
            retry_count += 1
            logger.debug(
                "%s is trying to find the socket file: %s (%d/%d)",
                self.name,
                self.protocol.addr,
                retry_count,
                CONN_MAX_RETRY,
            )
            time.sleep(CONN_CHECK_INTERVAL)
            continue

        logger.error(
            "%s cannot find the socket file: %s, will shutdown",
            self.name,
            self.protocol.addr,
        )

    def streaming(self):
        """Send stream data from the worker to the server through the socket."""
        while not self.shutdown.is_set():
            try:
                text, index = self.stream.get(timeout=self.timeout)
                # encode the text with UTF-8
                payloads = (text.encode(),)
                ids = (self.current_ids[index],)
                self.protocol.send(
                    HTTPStautsCode.STREAM_EVENT, ids, (0,) * len(ids), payloads
                )
                self.semaphore.release()
            except queue.Empty:
                continue

    def warmup(self):
        """Warmup to allocate resources (useful for GPU workload)[Optional]."""
        need_warmup = self.worker.example or self.worker.multi_examples
        if self.shutdown.is_set() or not need_warmup:
            return
        try:
            # Prioritize multi_examples if both appear.
            if self.worker.multi_examples:
                num_eg = len(self.worker.multi_examples)
                for i, example in enumerate(self.worker.multi_examples):
                    self.worker.forward(example)
                    logger.debug(
                        "warming up... (%d/%d)",
                        i + 1,
                        num_eg,
                    )
            else:
                self.worker.forward(self.worker.example)
            logger.info("%s warmup successfully", self.name)
        # pylint: disable=broad-except
        except Exception:
            logger.error(
                "%s warmup failed: %s, please ensure"
                " worker's example meets its forward input format",
                self.name,
                traceback.format_exc().replace("\n", " "),
            )

    def run(self):
        """Maintain the protocol connection and run the coordination."""
        while not self.shutdown.is_set():
            # reconnect if needed
            try:
                self.protocol.open()
            except OSError as err:
                if not self.shutdown_notify.is_set():
                    logger.error("%s socket connection error: %s", self.name, err)
                break

            self.coordinate()

    def decode(self, payload: bytes, state: int) -> Any:
        """Decode the payload with the state."""
        return (
            self.worker.deserialize(payload)
            if state & State.INGRESS
            else self.worker.deserialize_ipc(payload)
        )

    def encode(self, data: Any, state: int) -> bytes:
        """Encode the data with the state."""
        return (
            self.worker.serialize(data)
            if state & State.EGRESS
            else self.worker.serialize_ipc(data)
        )

    def coordinate(self):
        """Start coordinating the protocol's communication and worker's forward pass."""
        while not self.shutdown.is_set():
            try:
                # flag received from the server is not used
                _, ids, states, payloads = self.protocol.receive()
                # expose here to be used by stream event
                self.current_ids = ids
            except socket.timeout:
                continue
            except (struct.error, OSError) as err:
                if not self.shutdown_notify.is_set():
                    logger.error("%s failed to read from socket: %s", self.name, err)
                break

            # pylint: disable=broad-except
            length = len(payloads)
            try:
                data = [
                    self.decode(payload, state)
                    for (payload, state) in zip(payloads, states)
                ]
                with set_mosec_timeout(self.timeout):
                    data = (
                        self.worker.forward(data)
                        if self.worker.max_batch_size > 1
                        else (self.worker.forward(data[0]),)
                    )
                if len(data) != length:
                    raise ValueError(
                        "returned data size doesn't match the input data size:"
                        f"input({length})!=output({len(data)})"
                    )
                status = HTTPStautsCode.OK
                payloads = [
                    self.encode(datum, state) for (datum, state) in zip(data, states)
                ]
            except (MosecError, MosecTimeoutError) as err:
                err_msg = str(err).replace("\n", " - ")
                err_msg = err_msg if err_msg else err.msg
                logger.info("%s %s: %s", self.name, err.msg, err_msg)
                status = err.code
                payloads = [f"{err.msg}: {err_msg}".encode()] * length
            except Exception:
                logger.warning(traceback.format_exc().replace("\n", " "))
                status = HTTPStautsCode.INTERNAL_ERROR
                payloads = ["inference internal error".encode()] * length

            try:
                # pylint: disable=consider-using-with
                self.semaphore.acquire(timeout=self.timeout)
                self.protocol.send(status, ids, states, payloads)
            except OSError as err:
                logger.error("%s failed to send to socket: %s", self.name, err)
                break
            finally:
                self.semaphore.release()

        self.protocol.close()
        time.sleep(CONN_CHECK_INTERVAL)



# Copyright 2023 MOSEC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Environment variables related functions."""

from __future__ import annotations

import contextlib
import os
import warnings
from argparse import Namespace
from typing import Any, Dict, List, Union

MOSEC_ENV_PREFIX = "MOSEC_"
MOSEC_ENV_CONFIG = {
    "path": str,
    "capacity": int,
    "timeout": int,
    "address": str,
    "port": int,
    "namespace": str,
    "debug": bool,
    "dry_run": bool,
    "log_level": str,
}


@contextlib.contextmanager
def env_var_context(env: None | List[Dict[str, str]], index: int):
    """Manage the environment variables for a worker process."""
    default: Dict = {}
    try:
        if env is not None:
            for key, value in env[index].items():
                default[key] = os.getenv(key, "")
                os.environ[key] = value
        yield None
    finally:
        for key, value in default.items():
            os.environ[key] = value


def get_env_namespace(prefix: str = MOSEC_ENV_PREFIX) -> Namespace:
    """Read the config from environment variables before the argument parsing.

    Priority: CLI > env > default value.
    """
    namespace = Namespace()
    for name, converter in MOSEC_ENV_CONFIG.items():
        var = f"{prefix}{name.upper()}"
        value = os.environ.get(var)
        if not value:
            continue
        try:
            val = converter(value)
        except ValueError as err:
            warnings.warn(
                f"failed to convert env {var}={value} to type {converter} {err}, "
                "will skip this one",
            )
        else:
            setattr(namespace, name, val)

    return namespace


def validate_int_ge(number, name, threshold=1):
    """Validate int number is greater than threshold."""
    assert isinstance(
        number, int
    ), f"{name} must be integer but you give {type(number)}"
    assert number >= threshold, f"{name} must be no less than {threshold}"


def validate_str_dict(dictionary: Dict):
    """Validate keys and values of the dictionary is string type."""
    for key, value in dictionary.items():
        if not (isinstance(key, str) and isinstance(value, str)):
            return False
    return True


def validate_env(env: Union[Any, List[Dict[str, str]]], num: int):
    """Validate keys and values of the dictionary is string type."""
    if env is None:
        return
    assert len(env) == num, "len(env) must equal to num"
    valid = True
    if not isinstance(env, List):
        valid = False
    elif not all(isinstance(x, Dict) and validate_str_dict(x) for x in env):
        valid = False
    assert valid, "env must be a list of string dictionary"



# Copyright 2023 MOSEC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MOSEC type validation mixin."""

# pylint: disable=import-outside-toplevel

from typing import Any, Dict, Optional, Tuple

from mosec import get_logger
from mosec.errors import ValidationError
from mosec.utils import ParseTarget, parse_func_type
from mosec.worker import Worker

logger = get_logger()


class TypedMsgPackMixin(Worker):
    """Enable request type validation with `msgspec` and serde with `msgpack`."""

    # pylint: disable=no-self-use

    resp_mime_type = "application/msgpack"
    _input_typ: Optional[type] = None

    def deserialize(self, data: Any) -> Any:
        """Deserialize and validate request with msgspec."""
        import msgspec

        if not self._input_typ:
            self._input_typ = parse_func_type(self.forward, ParseTarget.INPUT)
        if not issubclass(self._input_typ, msgspec.Struct):
            # skip other annotation type
            return super().deserialize(data)

        try:
            return msgspec.msgpack.decode(data, type=self._input_typ)
        except msgspec.ValidationError as err:
            raise ValidationError(err) from err

    def serialize(self, data: Any) -> bytes:
        """Serialize with `msgpack`."""
        import msgspec

        return msgspec.msgpack.encode(data)

    @classmethod
    def get_forward_json_schema(
        cls, target: ParseTarget, ref_template: str
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """Get the JSON schema of the forward function."""
        import msgspec

        schema: Dict[str, Any]
        comp_schema: Dict[str, Any]
        schema, comp_schema = {}, {}
        typ = parse_func_type(cls.forward, target)
        try:
            (schema,), comp_schema = msgspec.json.schema_components(
                [typ], ref_template=ref_template
            )
        except TypeError as err:
            logger.warning(
                "Failed to generate JSON schema for %s: %s", cls.__name__, err
            )
        return schema, comp_schema



# Copyright 2023 MOSEC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dry run the service."""

from __future__ import annotations

import json
import signal
import sys
import time
from multiprocessing.context import SpawnContext, SpawnProcess
from typing import TYPE_CHECKING, Dict, List, Tuple, Union

from mosec.env import env_var_context
from mosec.log import get_internal_logger
from mosec.runtime import Runtime
from mosec.worker import Worker

if TYPE_CHECKING:
    from multiprocessing.connection import PipeConnection  # type: ignore
    from multiprocessing.synchronize import Event

logger = get_internal_logger()


def dry_run_func(
    worker_cls: type[Worker],
    batch: int,
    receiver: PipeConnection,
    sender: PipeConnection,
    ingress: bool,
    shutdown_notify: Event,
):
    """Dry run simulation function."""
    worker = worker_cls()
    while not shutdown_notify.is_set():
        if receiver.poll(timeout=0.1):
            break

    if shutdown_notify.is_set():
        return

    try:
        data = receiver.recv() if ingress else worker.deserialize(receiver.recv_bytes())
        logger.info("%s received %s", worker, data)
        data = worker.forward([data])[0] if batch > 1 else worker.forward(data)
        logger.info("%s inference result: %s", worker, data)
        data = worker.serialize(data)
        sender.send_bytes(data)
    # pylint: disable=broad-except
    except Exception as err:
        logger.error("get error in %s: %s", worker, err)
        shutdown_notify.set()


class Pool:
    """Process pool for dry run."""

    def __init__(self, process_context: SpawnContext, shutdown_notify: Event):
        """Initialize a process pool.

        Args:
            process_context: server context of spawn process
            shutdown_notify: event of server will shutdown
        """
        self.process_context = process_context
        self.shutdown_notify = shutdown_notify

        self.processes: List[SpawnProcess] = []
        self.sender_pipes: List[PipeConnection] = []
        self.receiver_pipes: List[PipeConnection] = []

    def new_pipe(self):
        """Create new pipe for dry run workers to communicate."""
        receiver, sender = self.process_context.Pipe(duplex=False)
        self.sender_pipes.append(sender)
        self.receiver_pipes.append(receiver)

    def start_worker(self, worker_runtime: Runtime, init: bool):
        """Start the worker process for dry run.

        Args:
            worker_runtime: worker runtime to start
            init: whether the worker is tried to start at the first time

        """
        self.new_pipe()
        coordinator = self.process_context.Process(
            target=dry_run_func,
            args=(
                worker_runtime.worker,
                worker_runtime.max_batch_size,
                self.receiver_pipes[-2],
                self.sender_pipes[-1],
                init,
                self.shutdown_notify,
            ),
            daemon=True,
        )

        with env_var_context(worker_runtime.env, 0):
            coordinator.start()

        self.processes.append(coordinator)

    def probe_worker_liveness(self) -> Tuple[Union[int, None], Union[int, None]]:
        """Check every worker is running/alive.

        Returns:
            index: index of the first failed worker
            exitcode: exitcode of the first failed worker

        """
        for i, process in enumerate(self.processes):
            if process.exitcode is not None:
                return i, process.exitcode
        return None, None

    def wait_all(self) -> Tuple[Union[int, None], Union[int, None]]:
        """Blocking until all worker to end or one failed.

        Returns:
            index: index of the first failed worker
            exitcode: exitcode of the first failed worker

        """
        for i, process in enumerate(self.processes):
            process.join()
            if process.exitcode != 0:
                return i, process.exitcode
        return None, None

    def first_last_pipe(self):
        """Get first sender and last receiver pipes."""
        return self.sender_pipes[0], self.receiver_pipes[-1]


class DryRunner:
    """Dry run the full stage.

    If examples are provided in the ingress :py:class:`Worker <mosec.worker.Worker>`,
    they will be used to pass through all the stages.

    For each stage, there will be only 1 worker. If `env` is provided during
    :py:meth:`append_worker <mosec.server.Server.append_worker>`, the 1st one
    will be used.
    """

    def __init__(self, router: Dict[str, List[Runtime]]):
        """Init dry runner."""
        self.router = router
        self.process_context: SpawnContext = SpawnContext()
        self.shutdown_notify: Event = self.process_context.Event()

        signal.signal(signal.SIGTERM, self.terminate)
        signal.signal(signal.SIGINT, self.terminate)

    def terminate(self, signum, framestack):
        """Terminate the dry run."""
        logger.info("received terminate signal [%s] %s", signum, framestack)
        self.shutdown_notify.set()

    def run(self):
        """Execute thr dry run process."""
        for endpoint, runtimes in self.router.items():
            logger.info(
                "init dry run for endpoint %s with %s",
                endpoint,
                [runtime.name for runtime in runtimes],
            )

            pool = Pool(self.process_context, self.shutdown_notify)
            pool.new_pipe()
            for i, worker_runtime in enumerate(runtimes):
                pool.start_worker(worker_runtime, i == 0)

            logger.info("dry run init successful")
            self.warmup(runtimes, pool)

            logger.info("wait for worker init done")
            if not self.shutdown_notify.is_set():
                self.shutdown_notify.set()

            failed, exitcode = pool.wait_all()
            if failed is not None:
                logger.warning(
                    "detect %s with abnormal exit code %d",
                    runtimes[failed].name,
                    exitcode,
                )
                sys.exit(exitcode)

            self.shutdown_notify.clear()
        logger.info("dry run exit")

    def warmup(self, runtimes: List[Runtime], pool: Pool):
        """Warmup the service.

        If neither `example` nor `multi_examples` is provided, it will only
        init the worker class.
        """
        ingress = runtimes[0].worker
        example = None
        if ingress.example:
            example = ingress.example
        elif ingress.multi_examples:
            assert isinstance(ingress.multi_examples, list), (
                "`multi_examples` " "should be a list of data"
            )
            example = ingress.multi_examples[0]

        if not example:
            logger.info("cannot find the example in the 1st stage worker, skip warmup")
            return

        sender, receiver = pool.first_last_pipe()
        start_time = time.perf_counter()
        sender.send(example)

        while not self.shutdown_notify.is_set():
            if receiver.poll(0.1):
                break
            # liveness probe
            failed, exitcode = pool.probe_worker_liveness()
            if failed is not None:
                logger.warning(
                    "worker %s exit with code %d",
                    runtimes[failed].name,
                    exitcode,
                )
                self.shutdown_notify.set()
                break

        if self.shutdown_notify.is_set():
            sys.exit(1)

        res = receiver.recv_bytes()
        duration = time.perf_counter() - start_time
        logger.info(
            "dry run result: %s",
            json.dumps(
                {
                    "request": example,
                    "result_size": len(res),
                    "warmup_duration": duration,
                }
            ),
        )



# Copyright 2023 MOSEC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MOSEC NumBin IPC worker mixin.

Features:

    * deserialize IPC data with numbin
    * serialize IPC data with numbin

Attention: numbin only supports NumPy ndarray types.
"""

# pylint: disable=import-outside-toplevel

from typing import Any

from mosec.errors import DecodingError, EncodingError


class NumBinIPCMixin:
    """NumBin IPC worker mixin interface."""

    # pylint: disable=no-self-use

    def serialize_ipc(self, data: Any) -> bytes:
        """Serialize with NumBin for the IPC."""
        import numbin

        try:
            data_bytes = numbin.dumps(data)
        except Exception as err:
            raise EncodingError from err
        return data_bytes

    def deserialize_ipc(self, data: bytes) -> Any:
        """Deserialize with NumBin for the IPC."""
        import numbin

        try:
            array = numbin.loads(data)
        except Exception as err:
            raise DecodingError from err
        return array



# Copyright 2022 MOSEC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MOSEC server interface.

This module provides a way to define the service components for machine learning
model serving.

Dynamic Batching
----------------

    The user may enable the dynamic batching feature for any stage when the
    corresponding worker is appended, by setting the
    :py:meth:`append_worker(max_batch_size) <Server.append_worker>`.

Multiprocessing
---------------

    The user may spawn multiple processes for any stage when the
    corresponding worker is appended, by setting the
    :py:meth:`append_worker(num) <Server.append_worker>`.
"""

import json
import multiprocessing as mp
import pathlib
import shutil
import signal
import subprocess
import traceback
from collections import defaultdict
from multiprocessing.synchronize import Event
from time import sleep
from typing import Dict, List, Type, Union

from mosec.args import parse_arguments
from mosec.dry_run import DryRunner
from mosec.log import get_internal_logger
from mosec.runtime import PyRuntimeManager, RsRuntimeManager, Runtime
from mosec.utils import ParseTarget
from mosec.worker import MOSEC_REF_TEMPLATE, SSEWorker, Worker

logger = get_internal_logger()


GUARD_CHECK_INTERVAL = 1
MOSEC_RESERVED_ENDPOINTS = {"/", "/metrics", "/openapi"}


class Server:
    """MOSEC server interface.

    It allows users to sequentially append workers they implemented, builds
    the workflow pipeline automatically and starts up the server.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(self):
        """Initialize a MOSEC Server."""
        self._shutdown: Event = mp.get_context("spawn").Event()
        self._shutdown_notify: Event = mp.get_context("spawn").Event()
        self._configs: dict = vars(parse_arguments())

        self._py_runtime_manager: PyRuntimeManager = PyRuntimeManager(
            self._configs["path"], self._shutdown, self._shutdown_notify
        )
        self._rs_runtime_manager = RsRuntimeManager(self._configs["timeout"])
        self._router: Dict[str, List[Runtime]] = defaultdict(list)

        self._daemon: Dict[str, Union[subprocess.Popen, mp.Process]] = {}

        self._server_shutdown: bool = False

    def _handle_signal(self):
        signal.signal(signal.SIGTERM, self._terminate)
        signal.signal(signal.SIGINT, self._terminate)

    def _validate_server(self):
        assert self._py_runtime_manager.worker_count > 0, (
            "no worker registered\n"
            "help: use `.append_worker(...)` to register at least one worker"
        )

    def _check_daemon(self):
        for name, proc in self._daemon.items():
            if proc is None:
                continue
            code = None
            if isinstance(proc, mp.Process):
                code = proc.exitcode
            elif isinstance(proc, subprocess.Popen):
                code = proc.poll()

            if code is not None:
                self._terminate(
                    code,
                    f"mosec daemon [{name}] exited on error code: {code}",
                )

    def _start_rs_runtime(self):
        """Subprocess to start the rust runtime manager program."""
        if self._server_shutdown:
            return

        # dump the config to a JSON file
        config_path = pathlib.Path(self._configs["path"]) / "config.json"
        configs = {"runtimes": [], "routes": []}
        for key, value in self._configs.items():
            if key in ("dry_run", "debug", "wait"):
                continue
            configs[key] = value
        for runtime in self._py_runtime_manager.runtimes:
            configs["runtimes"].append(
                {
                    "worker": runtime.name,
                    "max_batch_size": runtime.max_batch_size,
                    "max_wait_time": runtime.max_wait_time,
                }
            )
        for endpoint, pipeline in self._router.items():
            configs["routes"].append(
                {
                    "endpoint": endpoint,
                    "workers": [runtime.name for runtime in pipeline],
                    "is_sse": issubclass(pipeline[-1].worker, SSEWorker),
                    **generate_openapi([runtime.worker for runtime in pipeline]),
                }
            )
        config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(config_path, "w", encoding="utf-8") as file:
            json.dump(configs, file, indent=2)

        process = self._rs_runtime_manager.start(config_path)
        self.register_daemon("rs_runtime", process)

    def _terminate(self, signum, framestack):
        logger.info("received signum[%s], terminating server [%s]", signum, framestack)
        self._server_shutdown = True

    def _manage_py_runtime(self):
        init = True
        while not self._server_shutdown:
            failed_runtime = self._py_runtime_manager.check_and_start(init)
            if failed_runtime is not None:
                self._terminate(
                    1,
                    f"all the {failed_runtime.name} workers exited;"
                    " please check for bugs or socket connection issues",
                )
            init = False
            self._check_daemon()
            sleep(GUARD_CHECK_INTERVAL)

    def _halt(self):
        """Graceful shutdown."""
        # notify the rs runtime to shutdown
        self._shutdown_notify.set()
        self._rs_runtime_manager.halt()
        # shutdown py runtime manager
        self._shutdown.set()
        shutil.rmtree(self._configs["path"], ignore_errors=True)
        logger.info("mosec exited normally")

    def register_daemon(self, name: str, proc: subprocess.Popen):
        """Register a daemon to be monitored.

        Args:
            name: the name of this daemon
            proc: the process handle of the daemon
        """
        assert isinstance(name, str), "daemon name should be a string"
        assert isinstance(
            proc, (mp.Process, subprocess.Popen)
        ), f"{type(proc)} is not a process or subprocess"
        self._daemon[name] = proc

    # pylint: disable=too-many-arguments
    def append_worker(
        self,
        worker: Type[Worker],
        num: int = 1,
        max_batch_size: int = 1,
        max_wait_time: int = 0,
        start_method: str = "spawn",
        env: Union[None, List[Dict[str, str]]] = None,
        timeout: int = 0,
        route: Union[str, List[str]] = "/inference",
    ):
        """Sequentially appends workers to the workflow pipeline.

        Args:
            worker: the class you inherit from :class:`Worker<mosec.worker.Worker>`
                which implements the :py:meth:`forward<mosec.worker.Worker.forward>`
            num: the number of processes for parallel computing (>=1)
            max_batch_size: the maximum batch size allowed (>=1), will enable the
                dynamic batching if it > 1
            max_wait_time: the maximum wait time (millisecond) for dynamic batching,
                needs to be used with `max_batch_size` to enable the feature. If not
                configure, will use the CLI argument `--wait` (default=10ms)
            start_method: the process starting method ("spawn" or "fork"). (DO NOT
                change this unless you understand the difference between them)
            env: the environment variables to set before starting the process
            timeout: the timeout (second) for each worker forward processing (>=1)
            route: the route path for this worker. If not configured, will use the
                default route path `/inference`. If a list is provided, different
                route paths will share the same worker.
        """
        timeout = timeout if timeout >= 1 else self._configs["timeout"] // 1000
        max_wait_time = max_wait_time if max_wait_time >= 1 else self._configs["wait"]
        runtime = Runtime(
            worker,
            num,
            max_batch_size,
            max_wait_time,
            timeout,
            start_method,
            env,
        )
        self._register_route(runtime, route)
        self._py_runtime_manager.append(runtime)

    def register_runtime(self, routes: Dict[str, List[Runtime]]):
        """Register the runtime to the routes."""
        if self._py_runtime_manager.worker_count > 0:
            raise RuntimeError(
                "`register_runtime` can only be registered to an empty mosec server"
            )
        unique_runtimes = set()
        for endpoint, runtimes in routes.items():
            for runtime in runtimes:
                self._register_route(runtime, endpoint)
                unique_runtimes.add(runtime)
        for runtime in unique_runtimes:
            self._py_runtime_manager.append(runtime)

    def _register_route(self, runtime: Runtime, route: Union[str, List[str]]):
        """Register the route path for the worker."""
        if isinstance(route, str):
            if route in MOSEC_RESERVED_ENDPOINTS:
                raise ValueError(f"'{route}' is reserved, try another one")
            self._router[route].append(runtime)
        elif isinstance(route, list):
            for endpoint in route:
                if endpoint in MOSEC_RESERVED_ENDPOINTS:
                    raise ValueError(f"'{endpoint}' is reserved, try another one")
                self._router[endpoint].append(runtime)

    def run(self):
        """Start the mosec model server."""
        self._validate_server()
        if self._configs["dry_run"]:
            DryRunner(self._router).run()
            return

        self._handle_signal()
        self._start_rs_runtime()
        try:
            self._manage_py_runtime()
        # pylint: disable=broad-except
        except Exception:
            logger.error(traceback.format_exc().replace("\n", " "))
        self._halt()


def generate_openapi(workers: List[Type[Worker]]):
    """Generate the OpenAPI specification for one pipeline."""
    if not workers:
        return {}
    request_worker_cls, response_worker_cls = workers[0], workers[-1]
    input_schema, input_components = request_worker_cls.get_forward_json_schema(
        ParseTarget.INPUT, MOSEC_REF_TEMPLATE
    )
    return_schema, return_components = response_worker_cls.get_forward_json_schema(
        ParseTarget.RETURN, MOSEC_REF_TEMPLATE
    )

    def make_body(description, mime, schema):
        if not schema:
            return None
        return {"description": description, "content": {mime: {"schema": schema}}}

    return {
        "request_body": make_body(
            "Mosec Inference Request Body",
            request_worker_cls.resp_mime_type,
            input_schema,
        ),
        "responses": None
        if not return_schema
        else {
            "200": make_body(
                "Mosec Inference Result",
                response_worker_cls.resp_mime_type,
                return_schema,
            )
        },
        "schemas": {**input_components, **return_components},
        "mime": response_worker_cls.resp_mime_type,
    }



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "mosec"
copyright = "2023, mosec maintainers"
author = "mosec maintainers"
release = "latest"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "myst_parser",
    "sphinx_copybutton",
    "sphinxcontrib.programoutput",
    "sphinx_autodoc_typehints",
    "sphinxext.opengraph",
]

templates_path = ["_templates"]
exclude_patterns = []
source_suffix = [".rst", ".md"]
master_doc = "index"
language = "en"

# Extension configuration
myst_heading_anchors = 3
autodoc_member_order = "bysource"
# https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html
napoleon_attr_annotations = True
napoleon_include_init_with_doc = True
napoleon_use_admonition_for_references = True
# https://sphinxext-opengraph.readthedocs.io/en/latest/
ogp_site_url = "https://mosecorg.github.io/mosec/"
ogp_image = "https://user-images.githubusercontent.com/38581401/240117836-f06199ba-c80d-413a-9cb4-5adc76316bda.png"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_logo = "https://user-images.githubusercontent.com/38581401/240117836-f06199ba-c80d-413a-9cb4-5adc76316bda.png"
html_static_path = ["_static"]
html_favicon = "https://user-images.githubusercontent.com/38581401/134798617-0104dc12-e0d4-4ed5-a79c-9e2435e99a14.png"

# Theme
html_theme_options = {
    "sidebar_hide_name": True,
    "navigation_with_keys": True,
}



# Copyright 2023 MOSEC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Managers to control Coordinator and Mosec process."""

import multiprocessing as mp
import subprocess
from multiprocessing.context import ForkContext, SpawnContext
from multiprocessing.process import BaseProcess
from multiprocessing.synchronize import Event
from pathlib import Path
from time import monotonic, sleep
from typing import Callable, Dict, Iterable, List, Optional, Type, Union, cast

import pkg_resources

from mosec.coordinator import Coordinator
from mosec.env import env_var_context, validate_env, validate_int_ge
from mosec.log import get_internal_logger
from mosec.worker import Worker

logger = get_internal_logger()

NEW_PROCESS_METHOD = {"spawn", "fork"}


# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# pylint: disable=too-few-public-methods
class Runtime:
    """The wrapper with one worker and its arguments."""

    # count how many runtime instances have been created
    _stage_id: int = 0

    def __init__(
        self,
        worker: Type[Worker],
        num: int = 1,
        max_batch_size: int = 1,
        max_wait_time: int = 10,
        timeout: int = 3,
        start_method: str = "spawn",
        env: Union[None, List[Dict[str, str]]] = None,
    ):
        """Initialize the mosec coordinator.

        Args:
            worker (Worker): subclass of `mosec.Worker` implemented by users.
            num (int): number of workers
            max_batch_size: the maximum batch size allowed (>=1), will enable the
                dynamic batching if it > 1
            max_wait_time (int): the maximum wait time (millisecond)
                for dynamic batching, needs to be used with `max_batch_size`
                to enable the feature. If not configure, will use the CLI
                argument `--wait` (default=10ms)
            timeout (int): timeout (second) for the `forward` function.
            start_method: the process starting method ("spawn" or "fork")
            env: the environment variables to set before starting the process
        """
        self.worker = worker
        self.num = num
        self.max_batch_size = max_batch_size
        self.max_wait_time = max_wait_time
        self.timeout = timeout
        self.start_method = start_method
        self.env = env

        Runtime._stage_id += 1
        # adding the stage id in case the worker class is added to multiple stages
        self.name = f"{self.worker.__name__}_{self._stage_id}"
        self._pool: List[Union[BaseProcess, None]] = [None for _ in range(self.num)]

        self._validate()

    @staticmethod
    def _process_healthy(process: Union[BaseProcess, None]) -> bool:
        """Check if the child process is healthy.

        ref: https://docs.python.org/3/library/multiprocessing.html

            The exit code will be None if the process has not yet terminated.
        """
        return process is not None and process.exitcode is None

    def _healthy(self, method: Callable[[Iterable[object]], bool]) -> bool:
        """Check if all/any of the child processes are healthy."""
        return method(self._pool)

    def _start_process(
        self,
        worker_id: int,
        work_path: str,
        shutdown: Event,
        shutdown_notify: Event,
    ):
        """Start a worker process in the context."""
        context = mp.get_context(self.start_method)
        context = cast(Union[SpawnContext, ForkContext], context)
        coordinator_process = context.Process(
            target=Coordinator,
            args=(
                self.worker,
                self.max_batch_size,
                shutdown,
                shutdown_notify,
                work_path,
                self.name,
                worker_id + 1,
                self.timeout,
            ),
            daemon=True,
        )

        with env_var_context(self.env, worker_id):
            coordinator_process.start()

        self._pool[worker_id] = coordinator_process

    def _check(
        self,
        work_path: str,
        shutdown: Event,
        shutdown_notify: Event,
        init: bool,
    ) -> bool:
        """Check and start the worker process if it has not started yet.

        Args:
            work_path: path of working directory
            shutdown: Event of server shutdown
            shutdown_notify: Event of server will shutdown
            init: whether the worker is tried to start at the first time

        Returns:
            Whether the worker is started successfully
        """
        # for every sequential stage
        self._pool = [p if self._process_healthy(p) else None for p in self._pool]
        if self._healthy(all):
            # this stage is healthy
            return True
        if not init and not self._healthy(any):
            # this stage might contain bugs
            return False

        need_start_id = [
            worker_id for worker_id in range(self.num) if self._pool[worker_id] is None
        ]
        for worker_id in need_start_id:
            # for every worker in each stage
            self._start_process(worker_id, work_path, shutdown, shutdown_notify)
        return True

    def _validate(self):
        """Validate arguments of worker runtime."""
        validate_env(self.env, self.num)
        assert issubclass(
            self.worker, Worker
        ), "worker must be inherited from mosec.Worker"
        validate_int_ge(self.num, "worker number")
        validate_int_ge(self.max_batch_size, "maximum batch size")
        validate_int_ge(self.max_wait_time, "maximum wait time", 0)
        validate_int_ge(self.timeout, "forward timeout", 0)
        assert (
            self.start_method in NEW_PROCESS_METHOD
        ), f"start method must be one of {NEW_PROCESS_METHOD}"


class PyRuntimeManager:
    """The manager to control coordinator process."""

    def __init__(self, work_path: str, shutdown: Event, shutdown_notify: Event):
        """Initialize a coordinator manager.

        Args:
            work_path: path of working directory
            shutdown: Event of server shutdown
            shutdown_notify: Event of server will shutdown
        """
        self.runtimes: List[Runtime] = []

        self._work_path = work_path
        self.shutdown = shutdown
        self.shutdown_notify = shutdown_notify

    @property
    def worker_count(self) -> int:
        """Get number of workers."""
        return len(self.runtimes)

    @property
    def workers(self) -> List[Type[Worker]]:
        """Get List of workers."""
        return [r.worker for r in self.runtimes]

    def append(self, runtime: Runtime):
        """Sequentially appends workers to the workflow pipeline."""
        self.runtimes.append(runtime)

    def check_and_start(self, init: bool) -> Union[Runtime, None]:
        """Check all worker processes and try to start failed ones.

        Args:
            init: whether the worker is tried to start at the first time
        """
        for worker_runtime in self.runtimes:
            if not worker_runtime._check(  # pylint: disable=protected-access
                self._work_path, self.shutdown, self.shutdown_notify, init
            ):
                return worker_runtime
        return None


class RsRuntimeManager:
    """The manager to control Mosec process."""

    def __init__(self, timeout: int):
        """Initialize a Mosec manager.

        Args:
            timeout: service timeout in milliseconds
        """
        self.process: Optional[subprocess.Popen] = None

        self.server_path = Path(
            pkg_resources.resource_filename("mosec", "bin"), "mosec"
        )
        self.timeout = timeout

    def halt(self):
        """Graceful shutdown."""
        # terminate controller first and wait for a graceful period
        if self.process is None:
            return
        self.process.terminate()
        graceful_period = monotonic() + self.timeout / 1000
        while monotonic() < graceful_period:
            ctr_exitcode = self.process.poll()
            if ctr_exitcode is None:
                sleep(0.1)
                continue
            # exited
            if ctr_exitcode:  # on error
                logger.error("mosec service halted on error [%d]", ctr_exitcode)
            else:
                logger.info("mosec service halted normally [%d]", ctr_exitcode)
            break

        if monotonic() > graceful_period:
            logger.error("failed to terminate mosec service, will try to kill it")
            self.process.kill()

    def start(self, config_path: Path) -> subprocess.Popen:
        """Start the Mosec process.

        Args:
            config_path: configuration path of mosec
        """
        # pylint: disable=consider-using-with
        self.process = subprocess.Popen([self.server_path, config_path])
        return self.process



# Copyright 2022 MOSEC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command line arguments.

Arguments parsing for two parts:

    * prepared for the Rust service part
    * consumed by the Python worker part
"""

import argparse
import errno
import logging
import os
import random
import socket
import tempfile
import warnings

from mosec.env import get_env_namespace

DEFAULT_WAIT_MS = 10


def is_port_available(addr: str, port: int) -> bool:
    """Check if the port is available to use."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    err = sock.connect_ex((addr, port))
    sock.close()
    # https://docs.python.org/3/library/errno.html
    if err == 0:
        return False
    if err == errno.ECONNREFUSED:
        return True
    raise RuntimeError(
        f"Check {addr}:{port} socket connection err: {err}{errno.errorcode[err]}"
    )


def build_arguments_parser() -> argparse.ArgumentParser:
    """Build CLI arguments."""
    parser = argparse.ArgumentParser(
        description="Mosec Server Configurations",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="The following arguments can be set through environment variables: ("
        "path, capacity, timeout, address, port, namespace, debug, log_level, dry_run"
        "). Note that the environment variable should start with `MOSEC_` with upper "
        "case. For example: `MOSEC_PORT=8080 MOSEC_TIMEOUT=5000 python main.py`.",
    )

    parser.add_argument(
        "--path",
        help=(
            "Unix Domain Socket address for internal Inter-Process Communication."
            "If not set, a random path will be created under the temporary dir."
        ),
        type=str,
        default=os.path.join(
            tempfile.gettempdir(), f"mosec_{random.randrange(2**32):x}"
        ),
    )

    parser.add_argument(
        "--capacity",
        help="Capacity of the request queue, beyond which new requests will be "
        "rejected with status 429",
        type=int,
        default=1024,
    )

    parser.add_argument(
        "--timeout",
        help="Service timeout for one request (milliseconds)",
        type=int,
        default=3000,
    )

    parser.add_argument(
        "--wait",
        help="[deprecated] Wait time for the batcher to batch (milliseconds)",
        type=int,
        default=DEFAULT_WAIT_MS,
    )

    parser.add_argument(
        "--address",
        help="Address of the HTTP service",
        type=str,
        default="0.0.0.0",
    )

    parser.add_argument(
        "--port",
        help="Port of the HTTP service",
        type=int,
        default=8000,
    )

    parser.add_argument(
        "--namespace",
        help="Namespace for prometheus metrics",
        type=str,
        default="mosec_service",
    )

    parser.add_argument(
        "--debug",
        help="Enable the service debug log",
        action="store_true",
    )

    parser.add_argument(
        "--log-level",
        help="Configure the service log level",
        choices=["debug", "info", "warning", "error"],
        default="info",
    )

    parser.add_argument(
        "--dry-run",
        help="Dry run the service with provided warmup examples (if any). "
        "This will omit the worker number for each stage.",
        action="store_true",
    )
    return parser


def parse_arguments() -> argparse.Namespace:
    """Parse user configurations."""
    parser = build_arguments_parser()
    args, _ = parser.parse_known_args(namespace=get_env_namespace())

    if args.wait != DEFAULT_WAIT_MS:
        warnings.warn(
            "`--wait` is deprecated and will be removed in v1, please configure"
            "the `max_wait_time` on `Server.append_worker`",
            DeprecationWarning,
        )

    if args.debug:
        args.log_level = "debug"
        warnings.warn(
            "`--debug` is deprecated and will be removed in v1, please configure"
            "`--log_level=debug`",
            DeprecationWarning,
        )

    if not is_port_available(args.address, args.port):
        raise RuntimeError(
            f"{args.address}:{args.port} is in use. "
            "Please change to a free one (use `--port`)."
        )

    return args


def get_log_level() -> int:
    """Check if the service is running in debug mode."""
    parser = build_arguments_parser()
    args, _ = parser.parse_known_args(namespace=get_env_namespace())
    if args.debug:
        return logging.DEBUG

    level_map = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
    }
    return level_map[args.log_level]



# Copyright 2023 MOSEC Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provide useful utils to inspect function type."""

import inspect
from enum import Enum
from typing import List


class ParseTarget(Enum):
    """Enum to specify the target of parsing func type."""

    INPUT = "INPUT"
    RETURN = "RETURN"


def parse_func_type(func, target: ParseTarget) -> type:
    """Parse the input type of the target function.

    - single request: return the type
    - batch request: return the list item type
    """
    sig = inspect.signature(func)
    index = 0 if inspect.ismethod(func) else 1
    name = func.__name__
    if target == ParseTarget.INPUT:
        params = list(sig.parameters.values())
        if len(params) < index + 1:
            raise TypeError(
                f"`{name}` method doesn't have enough({index + 1}) parameters"
            )
        typ = params[index].annotation
    else:
        typ = sig.return_annotation

    origin = getattr(typ, "__origin__", None)
    if origin is None:
        return typ
    # GenericAlias, `func` could be batch inference
    if origin is list or origin is List:
        if not hasattr(typ, "__args__") or len(typ.__args__) != 1:  # type: ignore
            raise TypeError(
                f"`{name}` with dynamic batch should use "
                "`List[Struct]` as the input annotation"
            )
        return typ.__args__[0]  # type: ignore
    raise TypeError(f"unsupported type {typ}")
